from ...driver import Driver
from ...utils import async_cached_lock
from .apt import Apt
from .base import Installer
from .brew import Brew
from .dnf import DNF


async def linux_installer(driver: Driver) -> Installer:
    if await driver.executable_exists("dnf"):
        return DNF(driver)
    elif await driver.executable_exists("apt"):
        return Apt(driver)
    else:
        raise NotImplementedError("Cannot find a package manager.")


async def macos_installer(driver: Driver) -> Installer:
    return Brew(driver)


@async_cached_lock
async def make_installer(driver: Driver) -> Installer:
    os = await driver.os()
    installer_fn = os.switch(linux=linux_installer, macos=macos_installer)
    return await installer_fn(driver)



import subprocess
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Callable, Iterable
from contextlib import AbstractAsyncContextManager, asynccontextmanager, nullcontext
from dataclasses import dataclass
from functools import cache
from os import environ
from pathlib import Path
from typing import Any, Literal, Optional, cast

from trio import run_process

from .parallel import parallel_map_pause
from .utils import RunArg, Some, T, async_cached, intercalate, unsome


class OS(ABC):
    @abstractmethod
    def switch_(self, *, linux: Callable[["Linux"], T], macos: Callable[[], T]) -> T:
        pass

    def switch(self, *, linux: T, macos: T) -> T:
        return self.switch_(linux=lambda _: linux, macos=lambda: macos)


@dataclass
class Linux(OS):
    distribution: str

    def switch_(self, *, linux: Callable[["Linux"], T], macos: Callable[[], T]) -> T:
        return linux(self)

    RELEASE_FILE = "/etc/os-release"

    @classmethod
    async def get_distribution(cls, driver: "Driver") -> str:
        release_file = await driver.read_file(Path(cls.RELEASE_FILE))
        for line in release_file.splitlines():
            k, v = line.split("=", 1)
            if k == "ID":
                if v.startswith('"') and v.endswith('"'):
                    v = v[1:-1]
                return v

        raise ValueError(f"Cannot find distribution ID in {cls.RELEASE_FILE}.")


@dataclass
class MacOS(OS):
    def switch_(self, *, linux: Callable[["Linux"], T], macos: Callable[[], T]) -> T:
        return macos()


Architecture = Literal["x86_64", "aarch64"]


@dataclass
class RunResult:
    ok: bool


@dataclass
class RunResultOutput(RunResult):
    ok: bool
    output: str


class Driver(ABC):
    def __init__(self, *, root: bool = False) -> None:
        self.root = root
        super().__init__()

    def deconstruct(self) -> dict:
        return {"root": self.root}

    @cache
    def with_root(self, /, root: bool) -> "Driver":
        if self.root == root:
            return self
        kwargs = self.deconstruct() | {"root": root}
        return type(self)(**kwargs)

    @abstractmethod
    async def run_(
        self,
        *args: RunArg,
        check: bool = True,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        capture_output: bool = False,
        show_output: bool = False,
        silent: bool = False,
    ) -> RunResult:
        pass

    async def run(
        self,
        *args: RunArg,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        show_output: bool = False,
        silent: bool = False,
    ) -> None:
        await self.run_(*args, input=input, show_output=show_output, silent=silent)

    async def run_ok(self, *args: RunArg) -> bool:
        result = await self.run_(*args, check=False)
        return result.ok

    async def run_output(self, *args: RunArg, silent: bool = False) -> str:
        result = await self.run_(*args, capture_output=True, silent=silent)
        return cast(RunResultOutput, result).output

    async def run_output_(self, *args: RunArg) -> RunResultOutput:
        result = await self.run_(*args, capture_output=True, check=False, silent=True)
        return cast(RunResultOutput, result)

    async def executable_exists(self, executable: str) -> bool:
        return await self.run_ok("sh", "-c", f"command -v {executable}")

    async def find_executable(self, *executables: str) -> str:
        for candidate in executables:
            if await self.executable_exists(candidate):
                return candidate
        raise Exception(f"None of {', '.join(executables)} found in PATH.")

    async def is_file(self, path: Path) -> bool:
        return await self.run_ok("test", "-f", path)

    async def is_executable(self, path: Path) -> bool:
        return await self.run_ok("test", "-x", path)

    async def is_dir(self, path: Path) -> bool:
        return await self.run_ok("test", "-d", path)

    async def username(self) -> str:
        if self.root:
            return "root"
        return await self.run_output("whoami")

    @async_cached
    async def home(self) -> Path:
        path = "~root" if self.root else "~"
        result = await self.with_root(False).run_output("sh", "-c", f"eval echo {path}")
        return Path(result)

    async def local(self) -> Path:
        if self.root:
            return Path("/usr/local")
        return await self.home() / ".local"

    async def find(
        self,
        path: Path,
        *,
        name: Some[str] = None,
        file_type: Some[str] = None,
        maxdepth: Optional[int] = None,
    ) -> list[Path]:
        args: list[RunArg] = ["find", path, "-mindepth", "1"]

        def add_some_arg(arg: str, values: list[str]) -> None:
            if values:
                if len(values) > 1:
                    args.append("(")
                args.extend(intercalate("-o", ([arg, v] for v in values)))
                if len(values) > 1:
                    args.append(")")

        def add_optional_arg(arg: str, value: Optional[Any]) -> None:
            if value is not None:
                args.extend([arg, str(value)])

        add_optional_arg("-maxdepth", maxdepth)
        add_some_arg("-name", unsome(name))
        add_some_arg("-type", unsome(file_type))

        args.append("-print0")

        output = await self.run_output(*args)
        return [Path(result) for result in output.split("\0") if result]

    @asynccontextmanager
    async def tempfile(
        self, kind: Optional[Literal["directory"]] = None
    ) -> AsyncIterator[Path]:
        args = []
        if kind == "directory":
            args.append("-d")

        path = Path(await self.run_output("mktemp", *args))
        try:
            yield path
        finally:
            await self.rm(path)

    async def makedirs(self, path: Path) -> None:
        await self.run("mkdir", "-p", path)

    async def copy(self, source: Path, target: Path) -> None:
        await self.makedirs(target.parent)
        await self.rm(target)
        await self.run("cp", "-R", "-f", source, target)

    async def rm(self, path: Path) -> None:
        await self.run("rm", "-r", "-f", path)

    async def make_executable(self, path: Path) -> None:
        await self.run("chmod", "+x", path)

    async def read_file(self, path: Path) -> str:
        return await self.run_output("cat", path)

    async def write_file(self, path: Path, content: str) -> None:
        await self.makedirs(path.parent)
        await self.rm(path)
        await self.run("cp", "/dev/stdin", path, input=content.encode())

    async def link(
        self,
        source: Path,
        target: Path,
    ) -> None:
        await self.makedirs(target.parent)
        await self.rm(target)
        await self.run("ln", "-s", "-f", source, target)

    @async_cached
    async def os(self) -> OS:
        driver = self.with_root(False)
        os_type = await driver.run_output("uname")
        if os_type == "Linux":
            distribution = await Linux.get_distribution(driver)
            return Linux(distribution=distribution)
        elif os_type == "Darwin":
            return MacOS()
        else:
            raise ValueError(f"Unsupported OS type {os_type}.")  # pragma: no cover

    @async_cached
    async def architecture(self) -> Architecture:
        result = await self.with_root(False).run_output("uname", "-m")
        if result == "x86_64":
            return "x86_64"
        elif result in {"arm64", "aarch64"}:
            return "aarch64"
        else:
            raise ValueError(f"Unsupported architecture {result}.")  # pragma: no cover


class SubprocessDriver(Driver, ABC):
    def prepare_command(self, args: Iterable[RunArg]) -> list[RunArg]:
        return list(args)

    async def run_(
        self,
        *args: RunArg,
        check: bool = True,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        capture_output: bool = False,
        show_output: bool = False,
        silent: bool = False,
    ) -> RunResult:
        if capture_output and show_output:
            raise ValueError(
                "Cannot use capture_output and show_output at the same time."
            )  # pragma: no cover

        command = self.prepare_command(args)

        if not check or silent:
            stderr = subprocess.DEVNULL
        else:
            stderr = None

        # https://github.com/python/mypy/issues/5512
        cm: AbstractAsyncContextManager
        if show_output:
            cm = parallel_map_pause()
        else:
            cm = nullcontext()
        async with cm:
            result = await run_process(
                command,
                check=check,
                stdin=input,
                capture_stdout=not show_output,
                stderr=stderr,
                **self.run_args(),
            )

        ok = result.returncode == 0
        if capture_output:
            output = result.stdout.decode().strip()
            return RunResultOutput(ok=ok, output=output)
        else:
            return RunResult(ok=ok)

    def run_args(self) -> dict[str, Any]:
        return {}


class LocalDriver(SubprocessDriver):
    def prepare_command(self, args: Iterable[RunArg]) -> list[RunArg]:
        if self.root:
            return super().prepare_command(["sudo", *args])
        else:
            return super().prepare_command(args)

    def run_args(self) -> dict[str, Any]:
        result = super().run_args()
        if "VIRTUAL_ENV" in environ:
            # mybox is running in a virtual environment (for testing or
            # development). Remove virtual environment from PATH so that any
            # pip and pipx commands run in the user environment (they will
            # fail otherwise).
            new_environment = environ.copy()
            virtual_env = new_environment.pop("VIRTUAL_ENV")
            new_environment["PATH"] = ":".join(
                segment
                for segment in new_environment["PATH"].split(":")
                if not segment.startswith(virtual_env)
            )
            result["env"] = new_environment
        return result

    async def run_(
        self,
        *args: RunArg,
        check: bool = True,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        capture_output: bool = False,
        show_output: bool = False,
        silent: bool = False,
    ) -> RunResult:
        if args and args[0] == "sudo":
            # If the sudo prompt is needed, authenticate the user first while
            # showing the prompt.
            await super().run("sudo", "-v", show_output=True)
        return await super().run_(
            *args,
            check=check,
            input=input,
            capture_output=capture_output,
            show_output=show_output,
            silent=silent,
        )



from typing import Optional

from pydantic import Field

from ..tracker import Tracker
from ..utils import RunArg, async_cached, repo_version
from .destination import Destination

DEFAULT_REMOTE = "origin"


class Clone(Destination):
    repo: str = Field(..., alias="clone")
    branch: Optional[str] = None

    def derive_name(self) -> str:
        return self.repo

    async def directory_exists(self) -> bool:
        return await self.driver.is_dir(await self.destination())

    @async_cached
    async def git_args(self) -> list[RunArg]:
        return ["git", "-C", await self.destination()]

    async def run_git(self, *args: RunArg) -> None:
        await self.driver.run(*await self.git_args(), *args)

    async def run_git_output(self, *args: RunArg) -> str:
        return await self.driver.run_output(*await self.git_args(), *args)

    async def local_version(self) -> Optional[str]:
        if not await self.directory_exists():
            return None
        return await self.run_git_output("rev-parse", "HEAD")

    @property
    def remote(self):
        if self.repo.startswith("https://") or "@" in self.repo:
            return self.repo
        return f"https://github.com/{self.repo}.git"

    async def get_remote_version(self) -> str:
        return await repo_version(self.remote, branch=self.branch)

    async def branch_name(self, ref: str) -> str:
        return await self.run_git_output("rev-parse", "--abbrev-ref", ref)

    async def install(self, *, tracker: Tracker) -> None:
        destination = await self.destination()

        await self.driver.makedirs(destination.parent)

        if not await self.directory_exists():
            await self.driver.run("git", "clone", self.remote, destination)

        await self.run_git("remote", "set-url", DEFAULT_REMOTE, self.remote)

        if self.branch:
            remote_branch = f"{DEFAULT_REMOTE}/{self.branch}"
        else:
            remote_branch = await self.branch_name(f"{DEFAULT_REMOTE}/HEAD")

        branch = remote_branch.split("/")[1]
        await self.run_git("fetch", "--no-tags", DEFAULT_REMOTE, branch)

        current_branch = await self.branch_name("HEAD")
        if current_branch != branch:
            await self.run_git("switch", branch)
        await self.run_git("reset", "--hard", remote_branch)

        tracker.track(destination, root=self.root)

        await super().install(tracker=tracker)



import json
from typing import Optional

from ...driver import RunResultOutput
from ...utils import async_cached, async_cached_lock
from .base import PackageCacheInstaller, PackageVersionInfo


class Brew(PackageCacheInstaller):
    async def tap(self, repo: str) -> None:
        await self.driver.run(await self.brew(), "tap", repo)

    async def tapped(self) -> set[str]:
        output = await self.driver.run_output(await self.brew(), "tap")
        return set(output.strip().splitlines())

    async def install(self, package: str) -> None:
        await self.driver.run(await self.brew(), "install", package)
        await super().install(package)

    async def upgrade(self, package: str) -> None:  # pragma: no cover
        await self.driver.run(await self.brew(), "upgrade", package)
        await super().upgrade(package)

    @async_cached
    async def brew(self) -> str:
        # https://docs.brew.sh/FAQ#why-is-the-default-installation-prefix-opthomebrew-on-apple-silicon
        return await self.driver.find_executable("brew", "/opt/homebrew/bin/brew")

    @async_cached_lock
    async def brew_update(self) -> None:
        await self.driver.run(await self.brew(), "update")

    async def brew_info(self, *args: str) -> RunResultOutput:
        return await self.driver.run_output_(
            await self.brew(), "info", "--json=v2", *args
        )

    async def get_package_info(
        self, package: Optional[str]
    ) -> dict[str, PackageVersionInfo]:
        await self.brew_update()

        brew_result: RunResultOutput
        if not package:
            brew_result = await self.brew_info("--installed")
        else:
            brew_result = await self.brew_info(package)

        results: dict[str, PackageVersionInfo] = {}

        if not brew_result.ok:
            return results
        info = json.loads(brew_result.output)

        for cask in info["casks"]:
            if cask["tap"] == "homebrew/cask":
                # Do not require prefixes for casks from homebrew/cask
                name = cask["token"]
            else:
                name = f"{cask['tap']}/{cask['token']}"
            results[name] = PackageVersionInfo(
                installed=cask["installed"], latest=cask["version"]
            )

        for formula in info["formulae"]:
            name = formula["name"]
            try:
                installed = formula["installed"][0]["version"]
            except IndexError:
                installed = None
            latest = self.formula_version(formula)

            if installed and name in results:
                # Cask with the same name exists. This should not happen.
                raise ValueError(
                    f"Found cask and formula with the same name: {name}."
                )  # pragma: no cover
            results[name] = PackageVersionInfo(installed=installed, latest=latest)

        return results

    @staticmethod
    def formula_version(info: dict) -> str:
        version = info["versions"]["stable"]
        revision = info.get("revision")
        if revision:
            version += f"_{revision}"
        return version



import hashlib
from collections.abc import Iterator
from pathlib import Path
from typing import Optional

from pydantic import Field, field_validator

from ..tracker import Tracker
from ..utils import allow_singular
from .destination import Destination
from .manual_version import ManualVersion


class Links(ManualVersion, Destination):
    source: Path = Field(..., alias="links")

    @field_validator("source")
    @classmethod
    def source_absolute(cls, value):
        return Path(value).absolute()

    dot: bool = False
    shallow: bool = False
    only: Optional[list[str]] = None
    only_val = allow_singular("only")

    def derive_name(self) -> str:
        return f"links-{self.source}-{self.destination_}-{self.dot}-{self.root}"

    def all_paths(self) -> Iterator[Path]:
        if not self.source.is_dir():
            raise ValueError(f"Source is not a directory: {self.source}")

        if self.shallow:
            return (path for path in self.source.glob("*"))
        else:
            return (path for path in self.source.rglob("*") if path.is_file())

    def paths(self) -> Iterator[Path]:
        for path in self.all_paths():
            if self.only and path.name not in self.only:
                continue
            yield path

    async def get_remote_version(self) -> str:
        m = hashlib.sha256()
        for path in self.paths():
            m.update(str(path).encode())
        return m.hexdigest()

    async def install(self, *, tracker: Tracker) -> None:
        destination = await self.destination()

        for path in self.paths():
            target = path.relative_to(self.source)
            if self.dot:
                first_part, *parts = target.parts
                target = Path("." + first_part, *parts)
            target = destination.joinpath(target)

            await self.driver.link(path, target)
            tracker.track(target, root=self.root)

        await super().install(tracker=tracker)



import re
from abc import ABC, abstractmethod
from collections.abc import AsyncIterable, Iterable
from pathlib import Path

from pydantic import Field

from ..configparser import DesktopEntry
from ..tracker import Tracker
from ..utils import allow_singular_none
from .base import Package
from .manual_version import ManualVersion
from .system import SystemPackage

RESOLUTION_RE = re.compile(r"(\d+)x(\d+)")


class ManualPackage(ManualVersion, ABC):
    binaries: list[str] = Field(default_factory=list, alias="binary")
    binaries_val = allow_singular_none("binaries")

    binary_wrapper: bool = False

    apps: list[str] = Field(default_factory=list, alias="app")
    apps_val = allow_singular_none("apps")

    fonts: list[str] = Field(default_factory=list, alias="font")
    fonts_val = allow_singular_none("fonts")

    @abstractmethod
    async def binary_path(self, binary: str) -> Path:
        pass

    async def install_binary(self, name: str, tracker: Tracker) -> None:
        binary = await self.binary_path(name)
        target = await self.driver.local() / "bin" / name
        if self.binary_wrapper:
            await self.install_binary_wrapper(binary, target)
        else:
            await self.driver.link(binary, target)
        tracker.track(target, root=self.root)

    async def install_binary_wrapper(self, binary: Path, target: Path) -> None:
        await self.driver.write_file(target, f'#!/bin/sh\nexec "{binary}" "$@"')
        await self.driver.make_executable(target)

    @abstractmethod
    async def freedesktop_app_path(self, name: str) -> Path:
        pass

    @abstractmethod
    async def icon_paths(self, name: str) -> Iterable[Path]:
        pass

    async def icon_target_path(self, path: Path) -> Path:
        resolution: str
        if path.suffix == ".svg":
            resolution = "scalable"
        elif path.suffix == ".png":
            for part in path.parts:
                if RESOLUTION_RE.match(part):
                    resolution = part
                    break
            else:
                # Give a low resolution to prefer ones given explicitly
                resolution = "16x16"
        else:
            raise ValueError(f"Unknown icon type: '{path}'")

        return (
            (await self.driver.local())
            / "share"
            / "icons"
            / "hicolor"
            / resolution
            / "apps"
            / path.name
        )

    async def install_app(self, name: str, tracker: Tracker) -> None:
        await (await self.driver.os()).switch(
            linux=self.install_app_linux, macos=self.install_app_macos
        )(name, tracker)

    async def application_path(self) -> Path:
        return await self.driver.local() / "share" / "applications"

    async def install_desktop_file(self, path: Path, tracker: Tracker) -> None:
        target = await self.application_path() / path.name
        await self.driver.link(path, target)
        tracker.track(target, root=self.root)
        desktop_entry = DesktopEntry.from_string(await self.driver.read_file(target))
        if desktop_entry.icon:
            await self.install_icon(desktop_entry.icon, tracker)

    async def install_icon(self, icon: str, tracker: Tracker) -> None:
        for icon_path in await self.icon_paths(icon):
            target = await self.icon_target_path(icon_path)
            await self.driver.link(icon_path, target)
            tracker.track(target, root=self.root)

    async def install_app_linux(self, name: str, tracker: Tracker) -> None:
        path = await self.freedesktop_app_path(name)
        await self.install_desktop_file(path, tracker)

    async def install_app_macos(self, name: str, tracker: Tracker) -> None:
        # Installing macOS apps manually is not implemented.
        pass  # pragma: no cover

    @abstractmethod
    async def font_path(self, name: str) -> Path:
        pass

    async def install_font(self, name: str, tracker: Tracker) -> None:
        font_dir = (await self.driver.os()).switch(
            linux=await self.driver.local() / "share" / "fonts",
            macos=await self.driver.home() / "Library" / "Fonts",
        )
        source = await self.font_path(name)
        target = font_dir / source.name
        await self.driver.copy(source, target)
        tracker.track(target, root=self.root)
        await self.driver.run("fc-cache", "-f", font_dir)

    @abstractmethod
    async def install(self, *, tracker: Tracker) -> None:
        for binary in self.binaries:
            await self.install_binary(binary, tracker)
        for app in self.apps:
            await self.install_app(app, tracker)
        for font in self.fonts:
            await self.install_font(font, tracker)

        await super().install(tracker=tracker)

    async def prerequisites(self) -> AsyncIterable[Package]:
        async for package in super().prerequisites():
            yield package  # pragma: no cover

        if self.fonts:
            yield SystemPackage(
                system="fontconfig",
                db=self.db,
                driver=self.driver_,
            )



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path

from .driver import Driver
from .state import DB, INSTALLED_FILES, InstalledFile, Storage


@asynccontextmanager
async def tracking(*, driver: Driver, db: DB) -> AsyncIterator["ManagerTracker"]:
    installed_files = INSTALLED_FILES(db)

    stale = set(installed_files.find())
    current: set[InstalledFile] = set()

    tracker = ManagerTracker(
        storage=installed_files,
        stale=stale,
        current=current,
    )

    try:
        yield tracker
    finally:
        current_paths = set(current_file.path for current_file in current)

        for stale_file in stale:
            if stale_file.path not in current_paths:
                await driver.with_root(stale_file.root).rm(stale_file.path_)
            installed_files.delete(package=stale_file.package, path=stale_file.path)


class Tracker(ABC):
    @abstractmethod
    def track(self, target: Path, *, root: bool = False) -> None:
        pass


@dataclass
class ManagerTracker:
    storage: Storage[InstalledFile]
    stale: set[InstalledFile]
    current: set[InstalledFile]

    def skip(self, package: str) -> None:
        intact = {
            installed_file
            for installed_file in self.stale
            if installed_file.package == package
        }
        self.stale -= intact
        self.current |= intact

    def for_package(self, package: str) -> Tracker:
        return PackageTracker(parent=self, name=package)


@dataclass
class PackageTracker(Tracker):
    parent: ManagerTracker
    name: str

    def track(self, target: Path, *, root: bool = False) -> None:
        installed = InstalledFile(path=str(target), package=self.name, root=root)
        if installed not in self.parent.current:
            self.parent.current.add(installed)
            if installed in self.parent.stale:
                self.parent.stale.remove(installed)
            else:
                self.parent.storage.append(installed)
